import datetime
import json
from xml.etree import ElementTree

from pypiwrap.objects import Project, PyPIFeed, Stats


def test_parse_pypi_project() -> None:
    with open("tests/data/pypi_flask.json") as fp:
        project = Project.from_json(json.load(fp))

        assert project.name == "Flask"
        assert project.file_urls[0].filename == "flask-3.1.0-py3-none-any.whl"
        assert project.file_urls[0].size.bytes == 102_979
        assert project.file_urls[0].upload_time_tz == datetime.datetime(
            2024, 11, 13, 18, 24, 36, 135982, tzinfo=datetime.timezone.utc
        )


def test_parse_pypi_stats() -> None:
    with open("tests/data/pypi_stats.json") as fp:
        stats = Stats.from_json(json.load(fp))

        assert stats.total_size.bytes == 24_847_924_802_983
        assert stats.top_packages["PySide2"].bytes == 20_246_295_405
        assert list(stats.top_packages)[-1] == "OpenVisus"


def test_parse_pypi_feeds() -> None:
    with open("tests/data/pypi_packages.xml") as fp:
        rss = ElementTree.fromstring(fp.read())
        channel = rss.find("channel")

        assert channel is not None

        feed = PyPIFeed.from_xml(channel)

        assert feed.title == "PyPI newest packages"
        assert feed.link == "https://pypi.org/"
        assert feed.items[2].guid == feed.items[2].link
        assert feed.items[2].published_raw == "Fri, 17 Jan 2025 21:48:37 GMT"
        assert feed.items[2].published == datetime.datetime(2025, 1, 17, 21, 48, 37)



# Package metadata
__name__ = "pypiwrap"
__author__ = "Ángel Carias"
__license__ = "MIT"
__version__ = "2.0.0"

# PyPI constants
PYPI_HOST = "https://pypi.org"
USER_AGENT = f"aescarias/pypiwrap {__version__}"

SUPPORTED_SIMPLE_VERSION = (1, 3)
SIMPLE_CONTENT_TYPE = "application/vnd.pypi.simple.v1+json"

INTEGRITY_CONTENT_TYPE = "application/vnd.pypi.integrity.v1+json"



from __future__ import annotations

import warnings
from xml.etree import ElementTree

import requests

from .consts import PYPI_HOST, SIMPLE_CONTENT_TYPE, SUPPORTED_SIMPLE_VERSION, USER_AGENT
from .exceptions import (
    ParseError,
    UnexpectedVersionWarning,
    UnsupportedVersionError,
    raise_for_status,
)
from .objects import IndexPage, Project, ProjectPage, PyPIFeed, Stats


class PyPIFeedClient:
    """Client for the PyPI RSS feeds.

    .. versionadded:: 2.0.0
    .. warning:: This client is only designed for hosts under the pypi.org domain.

    Arguments:
        host (:class:`str`, optional):
            The base URL of the PyPI feeds host. Defaults to https://pypi.org.
    """

    def __init__(self, host=PYPI_HOST) -> None:
        self.host = host
        self.rest = requests.Session()
        self.rest.headers["User-Agent"] = USER_AGENT

    def __enter__(self):
        return self

    def __exit__(self, *exc_args) -> None:
        self.rest.close()

    def _get_feed(self, url: str) -> PyPIFeed:
        response = self.rest.get(url)
        raise_for_status(response)

        rss = ElementTree.fromstring(response.text)

        channel = rss.find("channel")

        if channel is None:
            raise ParseError("Could not parse RSS feed.")

        return PyPIFeed.from_xml(channel)

    def get_newest_packages(self) -> PyPIFeed:
        """Gets the newest packages created on PyPI."""
        return self._get_feed(f"{self.host}/rss/packages.xml")

    def get_latest_updates(self) -> PyPIFeed:
        """Gets the latest updates for individual projects on PyPI."""
        return self._get_feed(f"{self.host}/rss/updates.xml")

    def get_latest_releases_for_project(self, name: str) -> PyPIFeed:
        """Gets the latest releases for a project ``name``."""
        return self._get_feed(f"{self.host}/rss/project/{name}/releases.xml")


class PyPIClient:
    """Client for the PyPI JSON and Stats API.

    .. warning:: This client is only designed for hosts under the pypi.org domain.

    Arguments:
        host (:class:`str`, optional):
            The base URL of the PyPI API host. Defaults to https://pypi.org.
    """

    def __init__(self, host=PYPI_HOST) -> None:
        self.host = host
        self.rest = requests.Session()
        self.rest.headers["User-Agent"] = USER_AGENT

    def __enter__(self):
        return self

    def __exit__(self, *exc_args) -> None:
        self.rest.close()

    def get_project(self, name: str, version: str | None = None) -> Project:
        """Gets information about a project or any of its releases.

        Arguments:
            name (:class:`str`):
                The name of the project

            version (:class:`str`, optional):
                A version of the project to fetch. If none specified,
                the latest will be fetched.
        """

        if version:
            response = self.rest.get(f"{self.host}/pypi/{name}/{version}/json")
        else:
            response = self.rest.get(f"{self.host}/pypi/{name}/json")

        raise_for_status(
            response, {404: f"Could not find project or release for '{name}'"}
        )

        return Project.from_json(response.json())

    def get_stats(self) -> Stats:
        """Gets statistics about PyPI."""

        response = self.rest.get(
            f"{self.host}/stats", headers={"Accept": "application/json"}
        )
        raise_for_status(response)

        return Stats.from_json(response.json())


class SimpleRepoClient:
    """Client for the PyPI Simple Repository API (version 1).

    The methods included will emit a :class:`~.exceptions.UnexpectedVersionWarning`
    warning if it receives a response with a minor version greater than what's supported.

    As per PEP 629, the client will throw a :class:`~.exceptions.UnsupportedVersionError`
    exception if it receives a response with a major version greater than what's supported.

    Arguments:
        host (:class:`str`, optional):
            The base URL of the Simple Repository API host. Defaults to https://pypi.org.
    """

    def __init__(self, host: str = PYPI_HOST) -> None:
        self.host = host
        self.rest = requests.Session()
        self.rest.headers["Accept"] = SIMPLE_CONTENT_TYPE
        self.rest.headers["User-Agent"] = USER_AGENT

    def __enter__(self):
        return self

    def __exit__(self, *exc_args) -> None:
        self.rest.close()

    def _verify_api_version(self, version: str) -> None:
        declared_major, declared_minor = [int(comp) for comp in version.split(".")]
        expected_major, expected_minor = SUPPORTED_SIMPLE_VERSION

        if declared_major > expected_major:
            raise UnsupportedVersionError(
                f"API response returned version {declared_major}.{declared_minor}, "
                f"expected major version {expected_major} or lower."
            )
        elif declared_major == expected_major and declared_minor > expected_minor:
            warnings.warn(
                f"API response returned version {declared_major}.{declared_minor}, "
                f"this version is not strictly supported (latest supported: "
                f"{expected_major}.{expected_minor}).",
                UnexpectedVersionWarning,
            )

    def get_index_page(self) -> IndexPage:
        """Gets the index page for this repository.

        .. warning::
            If you're using the PyPI host, the response returned by PyPI could
            take several seconds to parse. Please use this method sparingly.
        """

        response = self.rest.get(f"{self.host}/simple")
        raise_for_status(response)

        page = response.json()

        self._verify_api_version(page["meta"]["api-version"])
        return IndexPage.from_json(page)

    def get_project_page(self, project: str) -> ProjectPage:
        """Gets the project page for a given ``project``."""

        response = self.rest.get(f"{self.host}/simple/{project}")
        raise_for_status(response)

        page = response.json()

        self._verify_api_version(page["meta"]["api-version"])
        return ProjectPage.from_json(page)



from __future__ import annotations

from typing import Any

from ..utils import remove_additional


class APIObject:
    """A base API object designed for JSON."""

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> APIObject:
        return cls(**remove_additional(cls, data.copy()))

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _build_repr_string(self, *args, **kwargs) -> str:
        arg_string = " ".join(map(repr, args))
        kwarg_string = " ".join(f"{key}={value!r}" for key, value in kwargs.items())
        final_repr = (self.__class__.__name__, arg_string, kwarg_string)

        return "<" + " ".join(item for item in final_repr if item) + ">"



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from xml.etree.ElementTree import Element


@dataclass
class PyPIFeed:
    """A simplified RSS feed returned by PyPI."""

    title: str
    """The title of this feed."""

    link: str
    """A link to the resource being aggregated."""

    description: str
    """A description or summary of this feed."""

    language: str
    """The language of this feed."""

    items: list[PyPIFeedItem]
    """A list of items aggregated by this feed."""

    @classmethod
    def from_xml(cls, element: Element) -> PyPIFeed:
        return cls(
            title=element.findtext("title", ""),
            link=element.findtext("link", ""),
            description=element.findtext("description", ""),
            language=element.findtext("language", ""),
            items=[
                PyPIFeedItem.from_xml(item) for item in (element.findall("item") or [])
            ],
        )


@dataclass
class PyPIFeedItem:
    """A simplified RSS feed item returned by PyPI."""

    title: str
    """The title or topic of this item."""

    link: str
    """A link to the resource being described."""

    guid: str
    """A globally unique identifier for this resource.
    
    In PyPI's case, this is usually the same as :attr:`PyPIFeedItem.link`.
    """

    published_raw: str
    """If provided, a string representing the datetime this resource was published.
    
    This datetime string usually follows RFC 822.
    """

    description: str
    """A description or summary of this item."""

    author: str
    """The author of this item."""

    @classmethod
    def from_xml(cls, element: Element) -> PyPIFeedItem:
        return cls(
            title=element.findtext("title", ""),
            link=element.findtext("link", ""),
            guid=element.findtext("guid", ""),
            published_raw=element.findtext("pubDate", ""),
            description=element.findtext("description", ""),
            author=element.findtext("author", ""),
        )

    @property
    def published(self) -> datetime | None:
        """If provided, the datetime this resource was published."""

        if self.published_raw:
            return datetime.strptime(self.published_raw, "%a, %d %b %Y %H:%M:%S %Z")



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any

from ..utils import Size, iso_to_datetime, remove_additional
from .base import APIObject


@dataclass
class DistributionFile(APIObject):
    """A file representing a package distribution."""

    filename: str
    """The filename of this distribution."""

    url: str
    """The download URL for this file."""

    size: Size
    """The size of the distribution."""

    hashes: dict[str, str]
    """A mapping of hash names to hex encoded digests for this file."""

    upload_time: datetime | None = None
    """The upload time for this file."""

    requires_python: str | None = None
    """The version constraints for this file if specified.
    
    This is equivalent to the 'Requires-Python' key in the Core metadata spec.
    """

    core_metadata: bool | dict[str, str] | None = None
    """
    An indication of whether metadata is available for this file.

    - If a boolean, whether this file has an associated metadata file.
    - If a dictionary, a mapping of hash names to hex encoded digests of the metadata file.
    - If None, this file has no associated metadata.
    """

    dist_info_metadata: bool | dict[str, str] | None = None
    """
    Contains the same values as :attr:`DistributionFile.core_metadata`. 
    
    When available, prefer using ``core_metadata`` over this attribute.
    """

    provenance_url: str | None = None  # API: provenance
    """If available, a URL to the file's associated provenance.
    
    See https://peps.python.org/pep-0740/#provenance-objects for details.
    
    .. versionadded:: 2.0.0
    """

    has_gpg_sig: bool | None = None  # API: gpg_sig
    """Whether a GPG signature for this file exists. If none, this value is unknown.
    
    For PyPI/Warehouse, this will always return None as GPG signatures are no longer
    supported by the service.
    """

    yanked: bool | str | None = None
    """
    - If a boolean, whether the file was yanked.
    - If a non-empty string, why the package was yanked.
    """

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> DistributionFile:
        # Certain API attributes, like requires-python, must be converted
        # to snake_case before unpacking
        result = {key.replace("-", "_"): val for key, val in data.items()}

        result["has_gpg_sig"] = result.get("gpg_sig")
        result["size"] = Size.from_int(result["size"])

        # See https://peps.python.org/pep-0714/ for why this is here.
        result["dist_info_metadata"] = result.pop("data_dist_info_metadata", None)

        result["provenance_url"] = result.pop("provenance", None)

        if result.get("upload_time") is not None:
            result["upload_time"] = iso_to_datetime(result["upload_time"])

        return cls(**remove_additional(cls, result))

    @property
    def gpg_url(self) -> str | None:
        """If available, a URL containing the GPG signature for this file."""
        if self.has_gpg_sig:
            return self.url + ".asc"

    @property
    def metadata_url(self) -> str | None:
        """If available, a URL containing the metadata file."""
        if self.core_metadata or self.dist_info_metadata:
            return self.url + ".metadata"

    def __repr__(self) -> str:
        return self._build_repr_string(self.filename, size=self.size.si)


@dataclass
class Meta(APIObject):
    """Information about a response from the Simple Repository API.

    .. versionadded:: 2.0.0
    """

    api_version: str
    """The API version being implemented. 
    
    See https://peps.python.org/pep-0629/ for details.
    """

    tracks: list[str]
    """If a repository, a list of project/repository URLs being "tracked" by the 
    extending repository.
    
    See https://peps.python.org/pep-0708/#repository-tracks-metadata for details.
    """

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Meta:
        return Meta(api_version=data["api-version"], tracks=data.get("tracks", []))

    def __repr__(self) -> str:
        return self._build_repr_string(api_version=self.api_version)


@dataclass
class IndexPage(APIObject):
    """The index page of the Simple Repository API.

    .. versionadded:: 2.0.0
    """

    meta: Meta
    """Information about the response."""

    projects: list[str]
    """A list of projects in the index."""

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> IndexPage:
        return cls(
            meta=Meta.from_json(data["meta"]),
            projects=[proj["name"] for proj in data["projects"]],
        )

    def __repr__(self) -> str:
        return self._build_repr_string(self.meta.api_version)


@dataclass
class ProjectPage(APIObject):
    """A project page from the Simple Repository API."""

    meta: Meta
    """Information about the response.

    .. versionadded:: 2.0.0
    """

    name: str
    """The name of this project."""

    alternate_locations: list[str]
    """A list of alternate locations or namespaces for this project. 
    
    See https://peps.python.org/pep-0708/#alternate-locations-metadata for details.
    
    .. versionadded:: 2.0.0
    """

    versions: list[str]
    """A list of all available versions for this project."""

    files: list[DistributionFile]
    """A list of distribution files for this project."""

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> ProjectPage:
        files = [DistributionFile.from_json(pkg_file) for pkg_file in data["files"]]

        return cls(
            meta=Meta.from_json(data["meta"]),
            name=data["name"],
            alternate_locations=data.get("alternate-locations", []),
            versions=data["versions"],
            files=files,
        )

    def __repr__(self) -> str:
        return self._build_repr_string(self.name)



from pypiwrap.objects.pypi import Project, ReleaseFile, Stats, Vulnerability
from pypiwrap.objects.rss import PyPIFeed, PyPIFeedItem
from pypiwrap.objects.simple_repo import DistributionFile, IndexPage, Meta, ProjectPage

__all__ = (
    "Stats",
    "Project",
    "IndexPage",
    "Meta",
    "ReleaseFile",
    "Vulnerability",
    "DistributionFile",
    "ProjectPage",
    "PyPIFeed",
    "PyPIFeedItem",
)



import json

import pytest

from pypiwrap.client import SimpleRepoClient
from pypiwrap.exceptions import UnexpectedVersionWarning, UnsupportedVersionError
from pypiwrap.objects import IndexPage, ProjectPage


def test_parse_index_page() -> None:
    with open("tests/data/simple_repo_index_page.json") as fp:
        index_json = json.load(fp)

    page = IndexPage.from_json(index_json)

    assert page.meta.api_version == "1.3"


def test_parse_project_page() -> None:
    with open("tests/data/simple_repo_colorama_page.json") as fp:
        project_json = json.load(fp)

    page = ProjectPage.from_json(project_json)

    assert page.meta.api_version == "1.3"
    assert page.name == "colorama"
    assert page.files[4].yanked
    assert page.files[4].filename == "colorama-0.4.2-py2.py3-none-any.whl"


def test_check_version_mismatch() -> None:
    with SimpleRepoClient() as client:
        with pytest.warns(UnexpectedVersionWarning):
            client._verify_api_version("1.25")

        with pytest.raises(UnsupportedVersionError):
            client._verify_api_version("10.0")



"""pypiwrap is an API wrapper for the Python Package Index."""

from .client import PyPIClient, PyPIFeedClient, SimpleRepoClient
from .consts import __author__, __license__, __name__, __version__

__all__ = (
    "SimpleRepoClient",
    "PyPIClient",
    "PyPIFeedClient",
    "__name__",
    "__author__",
    "__version__",
    "__license__",
)



from __future__ import annotations

import dataclasses
from datetime import datetime
from typing import Any, Literal, NamedTuple

SI_SUFFIXES = ["B", "KB", "MB", "GB", "TB"]
IEC_SUFFIXES = ["B", "KiB", "MiB", "GiB", "TiB"]


class Size(NamedTuple):
    """A tuple that includes human-readable representations of a file size."""

    bytes: int
    """The size represented in bytes."""

    iec: str
    """The size represented in binary/IEC units (KiB, MiB)."""

    si: str
    """The size represented in decimal/SI units (KB, MB)."""

    @classmethod
    def from_int(cls, num: int) -> Size:
        return Size(
            num, iec=bytes_to_readable(num, "iec"), si=bytes_to_readable(num, "si")
        )

    def __int__(self) -> int:
        return self.bytes


def iso_to_datetime(iso: str) -> datetime:
    """Converts an ISO 8601 string to a datetime object."""

    try:
        return datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S.%f%z")
    except ValueError:
        # some datetimes do not include a microsecond component
        return datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S%z")


def bytes_to_readable(number: float, unit: Literal["si", "iec"] = "si") -> str:
    """Converts a number (in bytes) to a human-readable string representation.

    Arguments:
        number (:class:`float`):
            A value in bytes.

        unit (:class:`str`, optional):
            Units to use when representing the result. May be ``si`` for decimal units (MB)
            which is the default or ``iec`` for binary units (MiB).
    """
    if unit == "iec":
        suffixes = IEC_SUFFIXES
        step_unit = 1024
    else:
        suffixes = SI_SUFFIXES
        step_unit = 1000

    # last one is assumed to be greatest
    suffix = None
    for suffix in suffixes:
        if number < step_unit:
            break
        number /= step_unit

    return f"{number:.2f} {suffix or suffixes[-1]}"


def remove_additional(cls: type[Any], data: dict[str, Any]) -> dict[str, Any]:
    """Takes any dataclass ``cls`` and a dictionary ``data`` that can unpack to it,
    discards any keys in ``data`` that are not fields of the dataclass, and returns
    a dictionary that can safely unpack to the dataclass."""

    result = data.copy()
    names = [field.name for field in dataclasses.fields(cls)]

    for key in data:
        if key not in names:
            result.pop(key)

    return result



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any

from ..utils import Size, iso_to_datetime, remove_additional
from .base import APIObject


@dataclass
class Project(APIObject):
    """A PyPI project or release.

    This includes information about the project, this particular release, URLs to
    distribution files, and any reported vulnerabilities.
    """

    author: str | None
    """The author of this project."""

    author_email: str | None
    """The email or contact details of the project's author."""

    classifiers: list[str]
    """A list of PyPI classifiers for this project.
    
    See https://pypi.org/classifiers for a complete list of classifiers.
    """

    description: str
    """A description of the project."""

    description_content_type: str | None
    """The content type of the description if available.
    
    PyPI supports 3 content types: ``text/plain``, ``text/x-rst`` (reStructuredText), and
    ``text/markdown``. PyPI will default to RST if no content type is specified or to
    plain text if the content type is invalid.
    """

    dynamic: list[str]
    """A list of distribution metadata values marked as Dynamic.
    
    Dynamic values are values that are expected to be "filled in later" by build backends.
    See https://peps.python.org/pep-0643/ for details.
    
    .. versionadded:: 2.0.0
    """

    keywords: str | None
    """Keywords relating to this project.
    
    This value is usually a list of comma or space separated values. It is equivalent
    to the 'Keywords' field in the Core metadata spec.
    """

    license: str | None
    """Text indicating the license for this project."""

    license_expression: str | None
    """If present, a valid SPDX license expression.
    
    .. versionadded:: 2.0.0
    """

    license_files: list[str]
    """A list of license files attached to this project, if any.
    
    .. versionadded:: 2.0.0
    """

    maintainer: str | None
    """The project's maintainer."""

    maintainer_email: str | None
    """The email or contact details of the project's maintainer."""

    name: str
    """The name of the project."""

    platform: str | None
    """The platform target of this release, if any.
    
    Usually, this is only set for platforms not included in the PyPI classifiers.
    """

    project_urls: dict[str, str]
    """A mapping of labels to URLs relating to this project."""

    project_url: str
    """The PyPI URL for this project."""

    release_url: str
    """The PyPI URL relating to this specific release."""

    requires_dist: list[str]
    """A list of required distributions or dependencies specified according to PEP 508."""

    requires_python: str | None
    """The Python version required for this release."""

    provides_extra: list[str]
    """A list of optional or extra features provided by this package.
    
    See https://peps.python.org/pep-0566/ for details. Distribution extra names should be 
    valid Python identifiers as defined in https://peps.python.org/pep-0685/.
    
    .. versionadded:: 2.0.0
    """

    summary: str
    """A short summary of this project."""

    version: str
    """The version of this project."""

    yanked: bool
    """Whether this release was 'yanked' or removed from circulation."""

    yanked_reason: str | None
    """The reason this release was yanked if applicable."""

    file_urls: list[ReleaseFile]
    """A list of files for this release."""

    vulnerabilities: list[Vulnerability]
    """A list of vulnerabilities for this release, if any."""

    last_serial: int
    """The most recent serial ID number for this project."""

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Project:
        info = remove_additional(cls, data["info"].copy())

        info["requires_dist"] = info.get("requires_dist") or []
        info["provides_extra"] = info.get("provides_extra") or []
        info["dynamic"] = info.get("dynamic") or []
        info["license_files"] = info.get("license_files") or []

        vulns = list(map(Vulnerability.from_json, data["vulnerabilities"]))
        files = list(map(ReleaseFile.from_json, data["urls"]))

        return cls(
            **info,
            last_serial=data["last_serial"],
            vulnerabilities=vulns,
            file_urls=files,
        )

    def __repr__(self) -> str:
        return self._build_repr_string(
            self.name,
            summary=self.summary,
            version=self.version,
            project_url=self.project_url,
        )


@dataclass
class Vulnerability(APIObject):
    """A vulnerability in a PyPI project or release."""

    aliases: list[str]
    """The identifiers used to refer to this vulnerability."""

    details: str
    """Details about this vulnerability."""

    fixed_in: list[str]
    """A list of releases where this vulnerability was addressed."""

    id: str
    """An identifier for this vulnerability."""

    link: str
    """A URL where more information about this vulnerability is provided."""

    source: str
    """The source from where this vulnerability report was obtained."""

    summary: str | None
    """A short summary of this vulnerability if available."""

    withdrawn: datetime | None = None
    """The datetime this vulnerability was withdrawn if applicable."""

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Vulnerability:
        if data.get("withdrawn") is not None:
            data["withdrawn"] = iso_to_datetime(data["withdrawn"])

        return cls(**remove_additional(cls, data))

    def __repr__(self) -> str:
        return self._build_repr_string(
            id=self.id, source=self.source, withdrawn=self.withdrawn
        )


@dataclass
class ReleaseFile(APIObject):
    """A file part of a PyPI release."""

    digests: dict[str, str]
    """A mapping of hash names to hex encoded digests corresponding to this release file.

    Usually, the digests available are ``md5``, ``sha256``, and ``blake2b_256``. 
    The keys available should be members of :attr:`hashlib.algorithms_guaranteed`.
    """

    filename: str
    """The filename for this release file."""

    package_type: str  # API: packagetype
    """The package type of this file. 
    
    Package types are grouped into 'sdist' for source distributions and 'bdist_*' for
    built distributions.
    
    As of PEP 527 and PEP 715, PyPI only accepts 'sdist' and 'bdist_wheel'. Other bdist 
    variants may still be returned for older packages but they are considered legacy 
    and can no longer be uploaded to PyPI.
    """

    python_version: str
    """The Python version target for this file.
    
    This value is 'source' for source distributions and a version target following 
    PEP 425 for built distributions.
    """

    requires_python: str
    """The Python version constraints for this file. 
    
    This is equivalent to the 'Requires-Python' key in the Core metadata specification.
    """

    size: Size
    """The size of this release file."""

    upload_time: datetime
    """The time this file was uploaded on."""

    upload_time_tz: datetime  # API: upload_time_iso_8601
    """The time this file was uploaded on in a format compliant with ISO 8601."""

    url: str
    """The URL for this release file."""

    yanked: bool
    """Whether this package was yanked."""

    yanked_reason: str | None
    """Why the package was yanked if applicable."""

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> ReleaseFile:
        data = data.copy()

        # Converting values to appropriate types
        data["size"] = Size.from_int(data["size"])
        data["upload_time_iso_8601"] = iso_to_datetime(data["upload_time_iso_8601"])
        data["upload_time"] = datetime.fromisoformat(data["upload_time"])

        # Renaming values to appropriate
        data["upload_time_tz"] = data.pop("upload_time_iso_8601")
        data["package_type"] = data.pop("packagetype")

        return cls(**remove_additional(cls, data))

    def __repr__(self) -> str:
        return self._build_repr_string(
            self.filename, size=self.size.si, package_type=self.package_type
        )


@dataclass
class Stats(APIObject):
    """Statistics about PyPI."""

    total_size: Size
    """The total size of all packages on PyPI combined."""

    top_packages: dict[str, Size]
    """A mapping of top packages sorted by size."""

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Stats:
        sorted_packages = sorted(
            data["top_packages"].items(), key=lambda item: item[1]["size"]
        )

        top_packages = {
            name: Size.from_int(package["size"]) for name, package in sorted_packages
        }

        return cls(
            total_size=Size.from_int(data["total_packages_size"]),
            top_packages=top_packages,
        )

    def __repr__(self) -> str:
        return self._build_repr_string(total_size=self.total_size.si)



from __future__ import annotations

import requests


def raise_for_status(
    response: requests.Response, messages: dict[int, str] | None = None
) -> None:
    """Raises an exception based on the response's status code. If the status code is
    deemed OK, this will do nothing.

    Arguments:
        response (:class:`requests.Response`): The response itself.

        messages (:class:`dict[int, str]`, optional):
            An optional mapping of status codes to messages.
    """

    if response.ok:
        return

    if messages is None:
        messages = {}

    error_map = {404: NotFound}

    exc = error_map.get(response.status_code, ClientError)
    raise exc(response.status_code, messages.get(response.status_code, response.reason))


class ClientError(Exception):
    """Raised when an error occurs while performing a request."""

    def __init__(self, status: int, reason: str) -> None:
        self.status = status
        self.reason = reason

        super().__init__(f"{status}: {reason}")


class NotFound(ClientError):
    """Raised when a project, release, or file was not found."""

    pass


class UnsupportedVersionError(Exception):
    """Raised when a response specifies a major version greater than the one supported.

    .. versionadded:: 2.0.0
    """

    pass


class UnexpectedVersionWarning(UserWarning):
    """Emitted when a response specifies a minor version greater than the one expected.

    .. versionadded:: 2.0.0
    """

    pass


class ParseError(Exception):
    """Raised when an error occurs while parsing a response.

    .. versionadded:: 2.0.0
    """

    pass
